@@ -55,15 +55,13 @@
 //! performance, should not be too costly), then $S\_i^2/n\_i$ on 64-bit integer.
 //! For $Sig$, $(n S\_i^2) / n\_i$ can be computed on 128-bit, as well as $S^2$.
 
+use crate::ScalibError;
 use hytra::TrAdder;
-use indicatif::{ProgressBar, ProgressFinish, ProgressStyle};
 use itertools::izip;
 use ndarray::{s, Array1, Array2, Array3, ArrayView2, ArrayViewMut2, Axis, Zip};
 use num_traits::{Bounded, PrimInt, Signed, WrappingAdd, Zero};
 use rayon::prelude::*;
 use std::convert::TryInto;
-use std::thread;
-use std::time::Duration;
 
 pub trait NativeInt: PrimInt + Signed + WrappingAdd + Send + Sync {}
 impl<T: PrimInt + Signed + WrappingAdd + Send + Sync> NativeInt for T {}
@@ -149,41 +147,13 @@ where
     /// number of samples in a trace
     ns: usize,
     /// number of classes
-    nc: u16,
+    nc: u32,
     /// max sample bit width
     bit_width: u32,
     /// total number of accumulated traces
     tot_n_samples: u32,
 }
 
-#[derive(Debug)]
-pub enum SnrError {
-    TooManyTraces,
-    ClassOverflow,
-    ClassOutOfBound,
-}
-
-impl std::fmt::Display for SnrError {
-    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
-        match self {
-            Self::TooManyTraces => write!(f, "Number of traces accumulated exceeds 2^32"),
-            Self::ClassOverflow => write!(f,
-    "The sum of samples (for a variable value) might overflow (threshold is 2^32 or 2^64)."
-               ),
-            Self::ClassOutOfBound => write!(
-                f,
-                "A class value of a variable is larger than the given number of classes."
-            ),
-        }
-    }
-}
-
-impl std::error::Error for SnrError {
-    fn source(&self) -> Option<&(dyn std::error::Error + 'static)> {
-        None
-    }
-}
-
 impl<T> SNR<T>
 where
     T: SnrType<Sample = i16> + std::fmt::Debug,
@@ -195,6 +165,7 @@ where
     /// np: number of independent random variable for which SNR must be estimated
     pub fn new(nc: usize, ns: usize, np: usize) -> Self {
         let ns8 = if ns % 8 == 0 { ns / 8 } else { ns / 8 + 1 };
+        assert!(nc <= 1 << 16);
         SNR {
             sum: Array3::from_elem((ns8, np, nc), [Zero::zero(); 8]),
             sum_square: Array1::from_elem((ns8,), [0; 8]),
@@ -216,47 +187,15 @@ where
         &mut self,
         traces: ArrayView2<T::Sample>,
         y: ArrayView2<u16>,
-    ) -> Result<(), SnrError> {
-        let x = traces;
-
-        let acc: TrAdder<u64> = TrAdder::new();
-        let acc_ref = &acc;
-
+        config: &crate::Config,
+    ) -> Result<(), ScalibError> {
         let n_it = (self.sum.shape()[0] as u64 + 3) / 4;
-        let n_updates = x.shape()[0] as u64 * x.shape()[1] as u64 * self.np as u64;
-
-        // Display bar if about 8E9 updates
-        if n_updates > (1 << 33) {
-            crossbeam_utils::thread::scope(|s| {
-                // spawn progress bar thread
-                s.spawn(move |_| {
-                    let pb = ProgressBar::new(n_it);
-                    pb.set_style(
-                        ProgressStyle::default_spinner()
-                            .template("{msg} [{elapsed_precise}] [{bar:40.cyan/blue}] (ETA {eta})")
-                            .on_finish(ProgressFinish::AndClear),
-                    );
-                    pb.set_message("Update SNR...");
-                    let mut x = 0;
-                    while x < n_it {
-                        pb.set_position(x);
-                        thread::sleep(Duration::from_millis(50));
-                        x = acc_ref.get();
-                    }
-                    pb.finish_and_clear();
-                });
-
-                // spawn computing thread
-                let res: Result<(), SnrError> = s
-                    .spawn(move |_| self.update_internal(traces, y, acc_ref))
-                    .join()
-                    .unwrap();
-                res
-            })
-            .unwrap()
-        } else {
-            self.update_internal(traces, y, acc_ref)
-        }
+        crate::utils::with_progress(
+            |it_cnt| self.update_internal(traces, y, it_cnt),
+            n_it,
+            "Update SNR",
+            config,
+        )
     }
 
     #[inline(never)]
@@ -266,7 +205,7 @@ where
         traces: ArrayView2<T::Sample>,
         y: ArrayView2<u16>,
         acc_ref: &TrAdder<u64>,
-    ) -> Result<(), SnrError> {
+    ) -> Result<(), ScalibError> {
         assert_eq!(traces.shape()[0], y.shape()[1]);
         assert_eq!(traces.shape()[1], self.ns);
         assert_eq!(y.shape()[0], self.np);
@@ -274,19 +213,19 @@ where
         assert!(y.is_standard_layout());
         let n_traces: u32 = traces.shape()[0]
             .try_into()
-            .map_err(|_| SnrError::TooManyTraces)?;
+            .map_err(|_| ScalibError::SnrTooManyTraces)?;
         self.tot_n_samples = self
             .tot_n_samples
             .checked_add(n_traces)
-            .ok_or(SnrError::TooManyTraces)?;
+            .ok_or(ScalibError::SnrTooManyTraces)?;
         let mut max_n_samples: u32 = 0;
         let nc = self.nc;
         let np = self.np;
         izip!(self.n_samples.outer_iter_mut(), y.outer_iter()).try_for_each(
             |(mut n_samples, y)| {
                 y.into_iter().try_for_each(|y| {
-                    if *y >= nc {
-                        Err(SnrError::ClassOutOfBound)
+                    if u32::from(*y) >= nc {
+                        Err(ScalibError::SnrClassOutOfBound)
                     } else {
                         n_samples[*y as usize] += 1;
                         max_n_samples = std::cmp::max(max_n_samples, n_samples[*y as usize]);
@@ -367,7 +306,10 @@ where
         // max_val does not overflow since max_n_samples < 2^32 and self.bit_width < 16
         let max_val = (max_n_samples as i64) << self.bit_width;
         if max_val > T::acc2i64(T::SumAcc::max_value()) {
-            return Err(SnrError::ClassOverflow);
+            return Err(ScalibError::SnrClassOverflow {
+                leak_upper_bound: 1 << self.bit_width,
+                max_n_traces: max_n_samples as i64,
+            });
         }
         return Ok(());
     }
